﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MeshVR;
using static SuperController;

namespace AlienAxeMan
{
    internal class AAM_Handcuffs : MVRScript
    {
        #region PluginInfo    
        private string PLUGIN_LABEL = "AAM Handcuffs";
        private string PLUGIN_VERSION = "1.0";
        #endregion

        private bool m_cuffs_initialized = false;

        protected JSONStorableStringChooser m_atom_string_chooser;

        protected JSONStorableBool m_collision_enabled;

        protected JSONStorableFloat left_cuff_open_amount;
        protected JSONStorableFloat right_cuff_open_amount;
        protected JSONStorableFloat ScrubDistance;

        private float m_previous_value_1;
        private float m_previous_value_2;

        public Transform m_cuff_left;
        public Transform m_cuff_open_left;

        protected Atom m_connected_to_atom = null;
        private Rigidbody m_connected_rigid_body = null;
        public GameObject m_chain_end_object;
        public Transform m_cuff_right;
        public Transform m_cuff_open_right;

        public float m_distance;
        public float m_max_distance = 1.0f;

        public float m_follow_speed = 5f;

        private FreeControllerV3 m_chain_end;

        public override void Init()
        {
            pluginLabelJSON.val = PLUGIN_LABEL + " " + PLUGIN_VERSION;

            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError(nameof(AAM_Handcuffs) + ": Containing object is not a \"Custom Unity Asset\", which is the only supported object type for this script.");
                    return;
                }

                // make atom selector (for selecting which "Right Cuff" this "Left Cuff" will link to)
                m_atom_string_chooser = new JSONStorableStringChooser("m_connected_to_atom", SuperController.singleton.GetAtomUIDs(), null, "Connect to Right Cuff", SyncAtom);
                RegisterStringChooser(m_atom_string_chooser);
                UIDynamicPopup dp = CreateScrollablePopup(m_atom_string_chooser);
                // want to always resync the atom choices on opening popup since atoms can be added/removed
                dp.popup.onOpenPopupHandlers += SyncAtomChoices;

                m_collision_enabled = new JSONStorableBool("Collision Enabled", false, OnCollisionEnabled);
                RegisterBool(m_collision_enabled);
                CreateToggle(m_collision_enabled);

                // make slider for opening/closing "Left Cuff"
                left_cuff_open_amount = new JSONStorableFloat("Left Cuff Open", 0f, OnLeftCuffOpenSliderChanged, 0f, 1f);
                RegisterFloat(left_cuff_open_amount);
                CreateSlider(left_cuff_open_amount, false);

                // make slider for opening/closing "Right Cuff"
                right_cuff_open_amount = new JSONStorableFloat("Right Cuff Open", 0f, OnRightCuffSliderChanged, 0f, 1f);
                RegisterFloat(right_cuff_open_amount);
                CreateSlider(right_cuff_open_amount, false);

                // A tiny bit of help to give some info to your fellow creator and/or player
                JSONStorableString helpText = new JSONStorableString("About",
                    "<color=#000><size=35><b>About</b></size></color>\n\n" +
                    "<color=#333>" +
                    "The handcuffs can be animated to open and close using Acidbubbles\' Timeline.\n\n" +
                    "Cuff openness will be maintained across scene saving and loading.\n\n" +
                    "</color>");
                UIDynamic helpTextfield = CreateTextField(helpText, true);
                helpTextfield.height = 1100.0f;

                SyncAtomChoices();
            }
            catch (Exception e)
            {
                SuperController.LogError(nameof(AAM_Handcuffs) + ": Exception caught: " + e);
            }
        }

        private void Start()
        {
            CustomUnityAssetLoader customUnityAssetLoader = containingAtom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>();
            if (!customUnityAssetLoader.isAssetLoaded)
            {
                customUnityAssetLoader.RegisterAssetLoadedCallback(OnLeftCuffAssetLoaded);
            }

            // if there's no "Right Cuff" selected yet, and we have not yet provided connection directions
            if (m_atom_string_chooser.val == null)
            {
                // Provide direction for how to connect to a "Right Cuff"
                SuperController.LogMessage("You can open the \"" + PLUGIN_LABEL + "\" plugin on \"" + containingAtom.name +
                    "\" and select which \"Right Cuff\" you\'d like to connect.");
            }
        }

        // Update is called once per frame
        void Update()
        {

        }

        void OnDestroy()
        {
            SuperController.singleton.onSubSceneLoadedHandlers -= OnSubSceneLoaded;

            // TODO: Reset openness to fully closed?  Leaving openness as is, can drift the baseline when reloading this script currently, for example.

            //Remove all references
            left_cuff_open_amount = null;
            right_cuff_open_amount = null;
            m_cuff_open_left = null;
            m_cuff_open_right = null;

            // remove all objects that script created
            Resources.UnloadUnusedAssets();
        }

        public void OnCollisionEnabled(bool collision_enabled)
        {
            UpdateColliders();
        }

        private bool InitCuffs()
        {
            if (m_cuffs_initialized)
            {
                return true;
            }

            CustomUnityAssetLoader customUnityAssetLoader = containingAtom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>();
            if (!customUnityAssetLoader.isAssetLoaded)
            {
                customUnityAssetLoader.RegisterAssetLoadedCallback(OnLeftCuffAssetLoaded);
                return false;
            }

            Transform left_control = GetChildObjectByPartialName(containingAtom.reParentObject, "control");
            if (m_cuff_open_left == null && left_control != null)
            {
                m_cuff_open_left = left_control.Find("Base/Cuff/Open");
                if (m_cuff_open_left == null)
                {
                    SuperController.LogError(nameof(AAM_Handcuffs) + ": m_cuff_open_left could not be found");
                }
            }

            if (m_connected_rigid_body != null)
            {
                Transform right_control = GetParentObjectByName(m_connected_rigid_body.transform, "controlr");

                if (m_cuff_open_right == null && right_control != null)
                {
                    m_cuff_open_right = right_control.Find("Base/Cuff/Open");
                    if (m_cuff_open_right == null)
                    {
                        SuperController.LogError(nameof(AAM_Handcuffs) + ": m_cuff_open_right could not be found");
                    }
                }
            }

            UpdateColliders();

            if (m_cuff_open_left != null && m_cuff_open_right != null)
            {
                m_cuffs_initialized = true;
                return true;
            }

            return false;
        }

        // set the cuffs' colliders enabled state according to the enabled JSON value
        private void UpdateColliders()
        {
            var UpdateCollider = new Action<Transform>(transform =>
            {
                if (transform == null)
                {
                    return;
                }

                Collider[] colliders = transform.root.GetComponentsInChildren<Collider>();
                foreach (Collider collider in colliders)
                {
                    if (collider.name == "Cuff" || collider.name == "larm" || collider.name == "rarm")
                    {
                        if (collider.enabled != m_collision_enabled.val)
                        {
                            collider.enabled = m_collision_enabled.val;
                        }
                    }
                }
            });

            UpdateCollider(m_cuff_open_left);
            UpdateCollider(m_cuff_open_right);
        }


        // Recursive function searching for a transform containing the string set in objectNameSearchString
        // It is used the first time for the initialization of the script
        // This script NEEDS to use a string "containing" your original because the cloning of the original asset will add "Clone" behind it.
        private Transform GetChildObjectByPartialName(Transform parent, string rootName)
        {
            foreach (Transform child in parent)
            {
                if (child.name.Contains(rootName))
                {
                    return child;
                }
                else
                {
                    Transform childSearch = GetChildObjectByPartialName(child, rootName);
                    if (childSearch != null)
                        return childSearch;
                }
            }

            return null;
        }

        private Transform getChildRoot(Transform parent, string searchName)
        {
            foreach (Transform child in parent)
            {
                if (child.name == searchName)
                {
                    return child;
                }
                else
                {
                    Transform childSearch = getChildRoot(child, searchName);
                    if (childSearch != null)
                        return childSearch;
                }
            }

            return null;
        }

        void OnLeftCuffOpenSliderChanged(float rotation)
        {
            if (m_cuff_open_left == null)
                return;

            var Slider1 = left_cuff_open_amount.val * -180.0f;
            var a = -(Slider1 - m_previous_value_1);
            m_cuff_open_left.localRotation *= Quaternion.Euler(0f, 0f, a);
            m_previous_value_1 = left_cuff_open_amount.val * -180.0f;
        }

        void OnRightCuffSliderChanged(float rotation)
        {
            if (m_cuff_open_right == null)
                return;

            var Slider2 = right_cuff_open_amount.val * 180.0f;
            var b = -(Slider2 - m_previous_value_2);
            m_cuff_open_right.localRotation *= Quaternion.Euler(0f, 0f, b);
            m_previous_value_2 = right_cuff_open_amount.val * 180.0f;
        }

        List<Atom> GetCUAAtoms()
        {
            var list = new List<Atom>();

            foreach (Atom atom in GetSceneAtoms())
            {
                if (atom.type == "CustomUnityAsset")
                {
                    list.Add(atom);
                }
            }

            return list;
        }

        protected void SyncAtomChoices()
        {
            List<string> atomChoices = new List<string>();
            atomChoices.Add("None");

            foreach (Atom atom in GetCUAAtoms())
            {
                GameObject chain_end = FindChainEnd(atom);
                if (chain_end != null)
                {
                    atomChoices.Add(atom.uid);
                }
            }
            m_atom_string_chooser.choices = atomChoices;
        }

        void OnLeftCuffAssetLoaded()
        {
            SyncAtom(m_atom_string_chooser.val);
        }

        void OnRightCuffAssetLoaded()
        {            
            SyncAtom(m_atom_string_chooser.val);
        }

        void OnSubSceneLoaded(SubScene subscene)
        {
            // TODO: fix case where it's a copy of a subscene, so as to address the issue where the copy is incorrectly loaded as being connected to the original subscene's Right Cuff
            // - perhaps I can detect that this plugin is in a subscene and that the connected Right Cuff would have been too, and then auto-switcharoo it?
            // - Note that this problem arises even after deleting the original Subscene, as it's just any time there's a subscene with a different name (such as the postfix #2)

            SyncAtom(m_atom_string_chooser.val);

            // TODO: handle the case where multiple subscenes may be loaded, such that the one I care about may not be first (so we wouldn't want to remove our handler yet)
            SuperController.singleton.onSubSceneLoadedHandlers -= OnSubSceneLoaded;
        }

        protected void SyncAtom(string atomUID)
        {
            if (atomUID != null)
            {
                if (atomUID == "None" || (m_connected_to_atom != null && m_connected_to_atom.uid != atomUID))
                {
                    m_cuffs_initialized = false;

                    m_connected_to_atom = null;
                    m_connected_rigid_body = null;
                    m_chain_end_object = null;
                    m_cuff_right = null;
                    m_cuff_open_right = null;

                    // TODO: destroy created object(s)?
                    // TODO: reset Right Cuff openness, including the slider

                    if (atomUID == "None")
                    {
                        return;
                    }
                }

                m_connected_to_atom = SuperController.singleton.GetAtomByUid(atomUID);
                if (m_connected_to_atom == null)
                {
                    // This is hacky, but I don't yet see a great way to address subscene loading, where we get the script's chooser callback prior to the
                    // Right Cuff's Atom being even created enough to acquire the asset loader for that.
                    SuperController.singleton.onSubSceneLoadedHandlers += OnSubSceneLoaded;

                    //SuperController.LogError(nameof(AAM_Handcuffs) + ": m_connected_to_atom is null ");
                    return;
                }

                CustomUnityAssetLoader customUnityAssetLoader = m_connected_to_atom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>();
                if (!customUnityAssetLoader.isAssetLoaded)
                {
                    customUnityAssetLoader.RegisterAssetLoadedCallback(OnRightCuffAssetLoaded);
                    return;
                }

                if (!m_connected_to_atom.isActiveAndEnabled)
                {
                    return;
                }

                m_chain_end_object = FindChainEnd(m_connected_to_atom);
                if (m_chain_end_object == null)
                {
                    SuperController.LogError(nameof(AAM_Handcuffs) + ": Failed to find chain end within: " + atomUID);
                    return;
                }
                else if (m_chain_end_object.GetComponent<Rigidbody>())
                {
                    ConnectToRightCuffChainEndRigidbody(m_chain_end_object.GetComponent<Rigidbody>());
                }
                else
                {
                    SuperController.LogError(nameof(AAM_Handcuffs) + ": m_chain_end_object is not an Rigidbody");
                    return;
                }
                
                InitCuffs();
            }
            else
            {
                SuperController.LogError(nameof(AAM_Handcuffs) + ": SyncAtom string atomUID = null ");
                return;
            }

            return;
        }

        Transform GetParentObjectByName(Transform obj, string targetName)
        {
            Transform parent = obj.parent;
            while (parent != null)
            {
                if (parent.gameObject.name == targetName)
                {
                    return parent;
                }
                parent = parent.parent;
            }
            return null;
        }

        private GameObject FindChainEnd(Atom atom)
        {
            if (atom == null)
            {
                return null;
            }

            Rigidbody[] rbs = atom.reParentObject.GetComponentsInChildren<Rigidbody>();
            foreach (Rigidbody rb in rbs)
            {
                if (rb.name == "ChainEnd")
                {
                    return rb.gameObject;
                }
            }
            return null;

        }

        void ConnectToRightCuffChainEndRigidbody(Rigidbody right_cuff_chain_end_rigid_body)
        {
            // if we don't have a valid Atom and Rigidbody to connect to, that's a problem
            if (m_connected_to_atom == null || right_cuff_chain_end_rigid_body == null)
            {
                SuperController.LogError(nameof(AAM_Handcuffs) + ": Connection settings invalid");
                return;
            }

            // if we're already connected to this Rigidbody, there's nothing more to do
            if (m_connected_rigid_body != null && m_connected_rigid_body == right_cuff_chain_end_rigid_body)
            {
                return;
            }

            foreach (Rigidbody left_cuff_rigid_body in containingAtom.rigidbodies)
            {
                if (left_cuff_rigid_body.name == "object")
                {
                    ConfigurableJoint configurable_joint = left_cuff_rigid_body.gameObject.GetComponent<ConfigurableJoint>();
                    if (configurable_joint == null)
                    {
                        configurable_joint = left_cuff_rigid_body.gameObject.AddComponent<ConfigurableJoint>();
                    }

                    if (configurable_joint != null)
                    {
                        configurable_joint.connectedBody = right_cuff_chain_end_rigid_body;
                        m_connected_rigid_body = right_cuff_chain_end_rigid_body;
                        configurable_joint.xMotion = ConfigurableJointMotion.Locked;
                        configurable_joint.yMotion = ConfigurableJointMotion.Locked;
                        configurable_joint.zMotion = ConfigurableJointMotion.Locked;
                        configurable_joint.autoConfigureConnectedAnchor = false;
                        configurable_joint.connectedAnchor = new Vector3(0.0f, -0.05f, 0.0f);
                        left_cuff_rigid_body.mass = right_cuff_chain_end_rigid_body.mass;
                    }

                    return;
                }
            }

            SuperController.LogError(nameof(AAM_Handcuffs) + ": Failed to connect to Right Cuff\'s Rigidbody");
        }
    }
}